// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_core_util_Recycler_hpp
#define anna_core_util_Recycler_hpp

#include <list>
#include <map>

#include <anna/core/RuntimeException.hpp>
#include <anna/core/Allocator.hpp>

namespace anna {

/**
   Mantiene una lista de punteros que puede crecer dinamicamente, no obstante, siempre que sea posible
   intenta reusar punteros creados previamente.

   @param T Clase de la que mantener la lista de punteros pre-asignados.
   @param Allocator Clase encargada de reservar la memoria para los objetos T en el momento en que sea necesaria
   una nueva instancia.

   \warning no actua como clase segura en MT, ver #anna::SafeRecycler
*/
template < typename T, typename _Allocator = Allocator <T> > class Recycler {
public:
  typedef typename std::list <T*> container;
  typedef typename container::iterator iterator;
  typedef typename container::const_iterator const_iterator;

  typedef typename std::map <T*, iterator> random_container;
  typedef typename random_container::iterator random_iterator;
  typedef typename random_container::value_type random_item;

  /**
     Constructor.
     \param randomAccess Indicador que permite activar el uso de estructuras de datos adicionales
     que permite que los mtodos #find y #release realicen la bsqueda del objeto de forma optimizada.
     Se ha comprobado que si necesitamos tratar en torno a un centenar de instancias
     es ms eficiente no activar las estructuras para acceso directo, para ms objetos resulta
     imprescinble.
  */
  Recycler(const bool randomAccess = false) {
    a_size = 0;
    a_randomContainer = (randomAccess == true) ? new random_container : NULL;
  }

  /**
     Destructor.
  */
  virtual ~Recycler() {
    clear();    // Pasa todos los punteros a_holes

    for(iterator ii = a_holes.begin(), maxii = a_holes.end(); ii != maxii; ii ++) {
      _Allocator::destroy(data(ii));
    }

    a_holes.clear();
    delete a_randomContainer;
  }

  /**
     Devuelve el nmero de elementos realmente utilizados hasta ahora.
     @return El nmero de elementos realmente utilizados hasta ahora.
  */
  int getSize() const throw() { return a_size; }

  /**
     Devuelve el nmero de elementos realmente utilizados hasta ahora.
     @return El nmero de elementos realmente utilizados hasta ahora.
  */
  int size() const throw() { return a_size; }

  /**
     Devuelve un puntero de tipo T. Solo crearia una nueva instancia de la clase T si al invocar a este
     metoodo no existe ninguna otra instancia que se pueda reutilizar, en cuyo caso haria una nueva reserva.

     Cada una de las llamadas a este metodo debe tener su correspondiente llamada al metodo  #release cuando
     el puntero deje de ser util.

     @return Un puntero a una instancia de tipo T.
  */
  T* create()
  throw(RuntimeException) {
    T* result = NULL;

    if(a_holes.empty() == false)  {
      iterator top = a_holes.begin();
      result = data(top);
      a_objects.splice(a_objects.end(), a_holes, top);
    } else
      a_objects.push_back(result = _Allocator::create());

    if(a_randomContainer != NULL) {
      iterator ii = a_objects.end();
      ii --;
      a_randomContainer->insert(random_item(result, ii));
    }

    a_size ++;
    return result;
  }

  /**
     Devuelve el iterador que apunta al objeto recibido como parametro. Si el objeto no se encuentra dentro de la
     lista devolver #end ()
     \return el iterador que apunta al objeto recibido como parametro.
  */
  iterator find(T* t)
  throw(RuntimeException) {
    iterator result = end();

    if(a_randomContainer != NULL) {
      random_iterator jj = a_randomContainer->find(t);

      if(jj != a_randomContainer->end())
        result = the_iterator(jj);
    } else {
      for(iterator ii = begin(), maxii = end(); ii != maxii; ii ++) {
        if(data(ii) == t) {
          result = ii;
          break;
        }
      }
    }

    return result;
  }

  /**
     Libera el puntero recibido como parametro. No se libera fisicamente sino que se deja marcado como
     reusable.

     Si el puntero pasado como parametro no ha sido obtenido mediante el metodo #create los resultados
     no estan definidos.

     @param t Instancia de un puntero de tipo T obtenido a partir del metodo #create.
  */
  void release(T* t)
  throw() {
    if(t == NULL)
      return;

    iterator result = end();

    if(a_randomContainer != NULL) {
      random_iterator jj = a_randomContainer->find(t);

      if(jj != a_randomContainer->end()) {
        result = the_iterator(jj);
        a_randomContainer->erase(jj);
      }
    } else {
      for(iterator ii = begin(), maxii = end(); ii != maxii; ii ++) {
        if(data(ii) == t) {
          result = ii;
          break;
        }
      }
    }

    if(result == end())
      return;

    a_holes.splice(a_holes.end(), a_objects, result);

    if(a_size > 0)
      a_size --;
  }

  /**
     Libera el puntero asociado al iterador recibido como parametro.
     \param ii Instancia a liberar.
  */
  void release(iterator ii) throw() { release(data(ii));  }

  /**
     Libera el puntero recibido como parametro. No se libera fisicamente sino que se deja marcado como
     reusable.

     Si el puntero pasado como parametro no ha sido obtenido mediante el metodo #create los resultados
     no estan definidos.

     @param t Instancia de un puntero de tipo T obtenido a partir del metodo #create.
  */
  void release(const T* t) throw() { release(const_cast <T*>(t)); }

  /**
     Marca como disponibles todos los objetos contenidos en memoria.
  */
  void clear()
  throw() {
    a_holes.splice(a_holes.end(), a_objects);
    a_size = 0;

    if(a_randomContainer)
      a_randomContainer->clear();
  }

  /**
     Devuelve un iterator al primer elemento, activo, contenido en el reciclador.
     \return Un iterator al primer elemento, activo, contenido en el reciclador.
  */
  iterator begin() throw() { return a_objects.begin(); }

  /**
     Devuelve un iterator al primer elemento, activo, contenido en el reciclador.
     \return Un iterator al primer elemento, activo, contenido en el reciclador.
  */
  const_iterator begin() const throw() { return a_objects.begin(); }

  /**
     Devuelve un iterator al final de la lista de elementos activos en el reciclador.
     \return Un iterator al final de la lista de elementos activos en el reciclador.
  */
  iterator end() throw() { return a_objects.end(); }

  /**
     Devuelve un iterator al final de la lista de elementos activos en el reciclador.
     \return Un iterator al final de la lista de elementos activos en el reciclador.
  */
  const_iterator end() const throw() { return a_objects.end(); }

  /**
     Devuelve el objeto referenciado por el iterator recibido como parametro.
     \return El objeto referenciado por el iterator recibido como parametro.
  */
  static T* data(iterator ii) throw() { return *ii; }

  /**
     Devuelve el objeto referenciado por el iterator recibido como parametro.
     \return El objeto referenciado por el iterator recibido como parametro.
  */
  static const T* data(const_iterator ii) throw() { return *ii; }

private:
  container a_objects;
  container a_holes;
  random_container* a_randomContainer;

  // Recordar que el list<T>::size tiene una eficiencia de O(N), mientras
  // que nuestro size, debera ser O(1), por eso hay que llevar la cuenta "a mano".
  int a_size;

//   static T* random_data (random_iterator ii) throw () { return ii->first; }
  static iterator the_iterator(random_iterator ii) throw() { return ii->second; }
};

}

#endif

