// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_core_util_TextComposer_hpp
#define anna_core_util_TextComposer_hpp

#include <vector>

#include <anna/core/RuntimeException.hpp>
#include <anna/core/mt/Mutex.hpp>
#include <anna/core/mt/Guard.hpp>
#include <anna/core//DataBlock.hpp>

#include <anna/core/util/TextVariable.hpp>

namespace anna {

class TextManager;
class TextVariable;

/**
   Compositor de textos dinamicos.

   Antes de invocar al metodo TextComposer::apply para obtener la cadena resultante de
   sustituir los enginees dinamicos habra que establecer el valor de cada uno de
   estos.

   La indicacion de forma es igual a la usada en los metodos printf, sprintf, etc ...

   Por ejemplo para establecer los valores de un compositor con la expresion:
   "Error en el circuito ${circuit:%d} de la central ${central:%05d} procesando mensaje ${messageid:%s}", deberiamos tener
   un codigo similar al que presentamos a continuacion:

   \code
   util::TextComposer& ttcc = textManager.find (N);

   util::TextComposer::Protector textComposer (ttcc);
   textComposer ["circuit"] = a;
   textComposer ["central"] = b;
   textComposer ["messageid"] = c;

   // Una vez hecho esto podemos obtener el texto
   std::cout << textComposer.apply () << std::endl;
   \endcode

   Donde hemos hecho las siguientes suposiciones:
      \li \em textManager es un gestor de texto definido dentro del ambito del ejemplo.
      \li \em N es el numero del compositor que deberia ser creado previamente indicando
      la expresion usada en el ejemplo.
      \li \em a Es una variable de tipo entero.
      \li \em b Es una variable de tipo entero.
      \li \em c Es una variable de tipo cadena.
*/
class TextComposer : std::vector <TextVariable*> {
  typedef std::vector <TextVariable*> container;
  typedef container::iterator variable_iterator;
  typedef container::const_iterator const_variable_iterator;

public:

  /**
     Destructor.
  */
  ~TextComposer();

  TextVariable& operator[](const char* varName) throw(RuntimeException) {
    return *(find(varName, Exception::Mode::Throw));
  }

  /**
   * Mtodo que realiza la interpretacin del texto de esta instancia.
   * Si la instancia se crea mediante un #util::TextManager no es necesario invocar a este mtodo.
   */
  void initialize() throw(RuntimeException);

  /**
     Devuelve el identificador del compositor, que coincidira con el indicado a la hora
     de crear el compositor mediante el metodo TextManager::create.
     \return El identificador del compositor.
  */
  const int getId() const throw() { return a_id; }

  /**
     Devuelve el n de variables asociadas al texto.
     \return el n de variables asociadas al texto.
  */
  int getNumberOfVariables() const throw() { return container::size(); }

  /**
     Devuelve una cadena con la informacion relevante de este objeto.
     \return Una cadena con la informacion relevante de este objeto.
  */
  String asString() const throw();

protected:
  /**
     Constructor.
     \param id Identificador del compositor de texto.
     \param expression Expression asociada al compositor. El numero de variables que hay que
     establecer antes de invocar a apply dependera de este texto.
  */
  TextComposer(const int id, const char* expression) : a_id(id), a_expression(expression), a_prefix(NULL), a_buffer(true) {;}

  /**
     Crea el texto teniendo en cuenta la expression y el contenido de las  variables asociadas.
     \return El texto teniendo en cuenta la expression y el contenido de las  variables asociadas.
  */
  String apply() const throw(RuntimeException);

  /**
   * Devuelve la instancia de la variable a la que apunta el iterador.
   * return la instancia de la variable a la que apunta el iterador.
   */
  static TextVariable* textVariable(variable_iterator ii) throw() { return *ii; }

  /**
   * Devuelve la instancia de la variable a la que apunta el iterador.
   * return la instancia de la variable a la que apunta el iterador.
   */
  static const TextVariable* textVariable(const_variable_iterator ii) throw() { return *ii; }

private:
  const int a_id;
  const String a_expression;
  String* a_prefix;
  Mutex a_mutex;
  mutable DataBlock a_buffer;

  TextComposer(const TextComposer&);

  TextVariable* find(const char* name, const Exception::Mode::_v) throw(RuntimeException);
  void lock() throw(RuntimeException) { a_mutex.lock(); }
  void unlock() throw() { a_mutex.unlock(); }

  static TextVariable::Type::_v calculeType(const char* format) throw(RuntimeException);

  friend class TextManager;
//   friend class Guard <TextComposer>;
//   friend class Protector;
};

}

#endif

