// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_io_Directory_hpp
#define anna_io_Directory_hpp

#include <sys/types.h>
#include <regex.h>

#include <vector>
#include <string>

#include <anna/core/RuntimeException.hpp>

namespace anna {

namespace io  {


/**
   Abstraccion de un directorio de archivos sobre UNIX.

   Permite recorrer el contenido de un directorio de archivos mediante iteradores.

   Ejemplo de uso:

   \code

   #include <anna/io/io.hpp>

   void f ()
      throw (anna::RuntimeException)
   {
      Directory dir;

      dir.read (".", Directory::Mode::ShortPath);

      for (Directory::const_iterator ii = dir.begin (), maxii = dir.end (); ii != maxii; ii ++) {
         const std::string& entry = Directory::data (ii);

         .... procesa la entrada ...
      }
   }

   \endcode
*/
class Directory {
public:
  /**
     Formas de obtener los archivos de un directorio.
  */
  struct Mode {
    enum _v {
      ShortPath = 0, /**< Retorna solo el nombre del archivo. */
      FullPath /**<  Retorna el nombre completo del archivo. */
    };
  };

  typedef std::vector <std::string> Files;
  typedef Files::const_iterator const_iterator;

  /**
     Constructor
  */
  Directory();

  /**
     Destructor.
  */
  virtual ~Directory() { clearPattern(); }

  /**
     Establece un filtro con la expresion regular que deben cumplir los archivos de este directorio.
     \param expression Expresion regular que deben cumplir los archivos.
     \warning Este metodo debe ser invocado antes de #read.
  */
  void setPattern(const char* expression) throw(RuntimeException);

  /**
     Elimina cualquier otro filtro que pudiera haber sido establecido mediante #setPattern.
  */
  void clearPattern() throw() { setPattern(NULL); }

  /**
     Lee las entradas del directorio indicado como parametro.
     \param dirName Ruta al nombre de directorio.
     \param mode Forma de obtener el nombre del los archivos.
  */
  void read(const char* dirName, const Mode::_v mode) throw(RuntimeException);

  /**
     Lee las entradas del directorio indicado como parametro.
     \param dirName Ruta al nombre de directorio.
     \param mode Forma de obtener el nombre del los archivos.
  */
  void read(const std::string& dirName,  const Mode::_v mode) throw(RuntimeException)  { read(dirName.c_str(), mode); }

  /**
     Devuelve un iterador a la primera entrada del directorio.
     \return Un iterador a la primera entrada del directorio.
  */
  const_iterator begin() const throw() { return a_files.begin(); }

  /**
     Devuelve un iterador a la ultima entrada del directorio.
     \return Un iterador a la utlima entrada del directorio.
  */
  const_iterator end() const throw() { return a_files.end(); }

  /**
     Devuelve un iterador a la entrada del directorio que corresponde con el nombre indicado.
     \return  un iterador a la entrada del directorio que corresponde con el nombre indicado.
  */
  const_iterator find(const std::string& file) const throw();

  /**
     Devuelve la entrada referenciado por el iterador recibido.
     \return  La entrada referenciado por el iterador recibido.
  */
  static const std::string& data(const_iterator& ii) throw() { return *ii; }

private:
  regex_t a_regex;
  bool a_hasPattern;
  Files a_files;
};

}
}

#endif
