// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_timex_Context_hpp
#define anna_timex_Context_hpp

#include <string>
#include <map>

#include <anna/core/functions.hpp>
#include <anna/core/mt/SafeSortedVector.hpp>
#include <anna/core/mt/Guard.hpp>
#include <anna/core/tracing/Logger.hpp>

#include <anna/xml/Node.hpp>
#include <anna/xml/Attribute.hpp>

#include <anna/timex/Engine.hpp>
#include <anna/timex/TimeEventObserver.hpp>
#include <anna/timex/Transaction.hpp>

#include <anna/core/util/Millisecond.hpp>

namespace anna {

namespace timex {

/**
 * Automatiza todas las operaciones relacionadas con el control de las transaciones de nuestra aplicacin.
 *
 * \param Tid Indica el tipo de dato usado para ordenar las transaciones en el contexto. Es distinto del identificador
 * de transaccin. El identificador de transaccin facilita su localizacin en los niveles de plataforma, mientras que
 * este \em Tid facilita su localizacin a niveles de aplicacin. Habitualmente este \em Tid ser una cadena alfanumrica.
 *
 * \warning
 *    \li El contexto asociado a la anna::timex::Transaction no puede ser usado al nivel de aplicacin, ya que esta
 * clase lo usa internamente.
 *    \li No se puede establecer un observador para estas transaciones ya que de otro modo el contexto no podra
 * seguir la pista de las operaciones realizadas sobre ella.
 *
 * El tipo de dato Tid debe tener los operadores requeridos para actuar como clave de una map &lt;Tid, Transaction&gt;
*/
template <typename Tid> class Context : public timex::TimeEventObserver {
public:
  /**
   * Crea una transaccin con el identificador recibido como parmetro.
   * En caso de que el identificador ya est registrado en el contexto obtendremos una excepcin.
   *
   * \param tid Identificador de aplicacin de la transacin.
   * \param classType Tipo de transacin que deseamos crear. Este parmetro slo ser til en caso de que nuestra aplicacin
   * tenga que gestionar varios tipos de transaciones u operaciones.
   *
   * Este mtodo terminar llamando al mtodo #createTransaction.
   *
   * \return La nueva transacin que ya habr sido activada.
   */
  Transaction* open(const Tid& tid, const int classType = 0)
  throw(RuntimeException) {
    Transaction* result = NULL;
    {
      Guard guard(a_mutex, "timex::Context::open");
      transaction_iterator ii = a_transactions.find(tid);

      if(ii != a_transactions.end()) {
        std::string msg(transaction(ii)->asString());
        msg += " | Already in progress";
        throw RuntimeException(msg, ANNA_FILE_LOCATION);
      }

      if((result = createTransaction(classType)) == NULL) {
        std::string msg("timex::Context::createTransaction returned NULL | ClassType: ");
        msg += functions::asString(classType);
        throw RuntimeException(msg, ANNA_FILE_LOCATION);
      }

      result->setTimeout(a_timeout);
      result->setObserver(this);
      result->setControlPoint();
      result->setId(anna_ptrnumber_cast(result));
      // Copia en el contexto de la transaccin la clave usada.
      // Se usar para poder buscar por clave a la hora de liberar la transaccin.
      std::pair <transaction_iterator, bool> rr = a_transactions.insert(value_type(tid, result));
      result->setContext((void*) &rr.first->first);
    }

    try {
      a_timeController.activate(result);
    } catch(RuntimeException&) {
      releaseTransaction(result);
      throw;
    }

    LOGINFORMATION(
      std::string msg("timex::Context::open | ");
      msg += result->asString();
      Logger::information(msg, ANNA_FILE_LOCATION);
    );
    return result;
  }

  /**
   * Devuelve al instancia de la transacin correspondiente al identificador recibido como parmetro.
   * \param tid Identificador de aplicacin de la transacin.
   * \param emode Modo de actuar en caso de que no se encuentre la transaccin correspondiente al identificador.
   * \return La transacin correspondiente al identificador recibido. Puede ser NULL.
   */
  Transaction* find(const Tid& tid, const Exception::Mode::_v emode = Exception::Mode::Throw)
  throw(RuntimeException) {
    Guard guard(a_mutex, "timex::Context::find");
    transaction_iterator ii = a_transactions.find(tid);
    Transaction* result = (ii != a_transactions.end()) ? transaction(ii) : NULL;

    if(result == NULL) {
      if(Exception::Mode::Throw) {
        std::string msg("Transaction: ");
        msg += identifierAsString(tid);
        msg += " | Tid not found";
        throw RuntimeException(msg, ANNA_FILE_LOCATION);
      } else if(Exception::Mode::Trace && Logger::isActive(Logger::Warning)) {
        std::string msg("Transaction: ");
        msg += identifierAsString(tid);
        msg += " | Tid not found";
        Logger::warning(msg, ANNA_FILE_LOCATION);
      }
    }

    return result;
  }

  /**
   * Obtiene el identificador de aplicacin asociado a la transacin recibida como parmetro.
   * \param transaction Puntero a la transacin a procesar. Debe haber sido creada con el mtodo #open.
   * \return El identificador de aplicacin indicado en el mtodo #open.
   */
  const Tid& getIdentifier(const Transaction* transaction) const
  throw(RuntimeException) {
    if(transaction == NULL)
      throw RuntimeException("timex::Context::getIdentifier | Can not work with NULL transaction", ANNA_FILE_LOCATION);

    const void* context =  transaction->getContext();

    if(context == NULL) {
      std::string msg("timex::Context::getIdentifier | ");
      msg += transaction->asString();
      msg += " | Can not work with NULL context";
      throw RuntimeException(msg, ANNA_FILE_LOCATION);
    }

    return contextAsIdentifier(context);
  }

  /**
   * Termina la transacin antes de que termine su periodo de espera.
   * \param transaction Transacin a terminar. Si fuera NULL este mtodo no tiene ningn efecto.
   */
  void close(Transaction* transaction)
  throw() {
    if(transaction == NULL)
      return;

    LOGINFORMATION(
      std::string msg("timex::Context::close | ");
      msg += transaction->asString();
      Logger::information(msg, ANNA_FILE_LOCATION);
    );

    try {
      a_timeController.cancel(transaction);
    } catch(RuntimeException& ex) {
      ex.trace();
    }
  }

  /**
   * Devuelve la informacin relevante de esta clase en forma de documento XML.
   * \param parent Nodo XML del que colgar la informacin de esta instancia.
   * \return Un documento XML con la informacin relevante de esta clase.
   */
  virtual xml::Node* asXML(xml::Node* parent) const
  throw() {
    xml::Node* result = parent->createChild("timex.Context");
    result->createAttribute("Size", a_transactions.size());
    result->createAttribute("AvgSize", (a_counter == 0) ? 0 : a_accSize / a_counter);
    result->createChild("Timeout")->createAttribute("Millisecond", a_timeout.getValue());
    result->createChild("AvgResponseTime")->createAttribute("Millisecond", (a_counter == 0) ? 0 : a_accDelay / a_counter);
    return result;
  }

  /**
   * Obtiene la instancia del gestor del contexto asociado a la transacin recibida como parmetro.
   * Puede ser NULL.
   * \param transaction Puntero a la transacin a procesar. Si no ha sido creada con el mtodo #open
   * el resultado de esta operacin no est determinado.
   * \return la instancia del gestor del contexto asociado a la transacin recibida como parmetro.
   * Puede ser NULL.
   */
  static const Context* getInstance(const Transaction* transaction)
  throw(RuntimeException) {
    if(transaction == NULL)
      throw RuntimeException("timex::Context::getInstance | Can not work with NULL transaction", ANNA_FILE_LOCATION);

    return reinterpret_cast <const Context*>(transaction->getObserver());
  }

protected:
  /**
   * Constructor.
   *
   * \param observerName Esta clase hera de #anna::timex::TimeEventObserver por lo que debe tener un nombre lgico
   * \param timeController Gestor de temporizacin usado en la aplicacin.
   * \param timeout Duracin de las transaciones creadas por el contexto. En principio todas las transaciones tendrn la misma
   * duracin, pero en prximas versiones se podra ampliar la funcionalidad para que cada transacin pueda indicar su duracin de
   * forma independiente.
   */
  Context(const char* observerName, timex::Engine& timeController, const Millisecond& timeout) :
    TimeEventObserver(observerName),
    a_timeController(timeController),
    a_timeout(timeout) {
    a_counter = a_accDelay = a_accSize = 0;
  }

  /**
   * Mtodo virtual puro que crear las transaciones de este contexto cuando sea necesario.
   * Lo ms adecuado sera implementar este mtodo mediate el uso del patrn #anna::Recycler.
   * \param classType Tipo de transacin que deseamos crear. Este parmetro slo ser til en caso de que nuestra aplicacin
   * tenga que gestionar varios tipos de transaciones u operaciones.
   * \return Una nueva transacin que puede ser usada por este contexto.
   */
  virtual Transaction* createTransaction(const int classType) throw() = 0;

  /**
   * Mtodo virtual puro que liber el espacio asociado a la transacin recibida como parmetro.
   * Lo ms adecuado sera implementar este mtodo mediate el uso del patrn #anna::Recycler.
   * \param transaction Transacin a liberar.
   */
  virtual void releaseTransaction(Transaction* transaction) throw() = 0;

  /**
   * Mtodo virtual puro que debe devolver una \em std::string correspondiente al valor del identificador
   * recibido como parmetro.
   * \param tid Identificador nico que la transacin.
   * \return una \em std::string correspondiente al valor del identificador recibido como parmetro.
   */
  virtual std::string identifierAsString(const Tid& tid) const throw() = 0;

  /**
   * Mtodo virtual puro que debe devolver la clave de aplicacin asociada a una transacin
   * \param tid Puntero que apunta a la clave de aplicacin usada para acceder a una transaccin.
   * \return La clave de aplicacin correspondiente al puntero recibido como parmetro.
   */
  virtual const Tid& contextAsIdentifier(const void* tid) const throw() = 0;

private:
  typedef typename std::map <Tid, Transaction*> transaction_container;
  typedef typename transaction_container::iterator transaction_iterator;
  typedef typename transaction_container::value_type value_type;

  Engine& a_timeController;
  transaction_container a_transactions;
  Mutex a_mutex;
  Millisecond a_timeout;
  unsigned int a_counter;
  unsigned int a_accDelay;
  unsigned int a_accSize;

  /* Mtodo re-implementado de timex::TimeEventObserver */
  void release(timex::TimeEvent* event)
  throw() {
    Transaction* tt = static_cast <Transaction*>(event);
    Guard guard(a_mutex, "timex::Context::release");
    // Calcula la duracin media de las transaciones de este contexto.
    const Millisecond &delay = tt->getDelay();
    const int size = a_transactions.size();

    /* Si desborda el acumulador de tiempo ... */
    if(++ a_counter == 0 || (a_accDelay + delay) < a_accDelay || (a_accSize + size) < a_accSize) {
      a_counter = 1;
      a_accDelay = 0;
      a_accSize = 0;
    }

    const Tid& tid(contextAsIdentifier(tt->getContext()));

    if(a_transactions.erase(tid) == 0) {
      std::string msg("Transaction: ");
      msg += identifierAsString(tid);
      msg += " | Tid can not be erased from map";
      Logger::warning(msg, ANNA_FILE_LOCATION);
    }

    releaseTransaction(tt);
    a_accDelay += delay;
    a_accSize += a_transactions.size();
  }

  static Transaction* transaction(transaction_iterator ii) throw() { return ii->second; }
};

}
}

#endif


