// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <algorithm>

#include <anna/core/tracing/TraceMethod.hpp>
#include <anna/core/functions.hpp>
#include <anna/core/tracing/Logger.hpp>

#include <anna/xml/Node.hpp>

#include <anna/comm/Delivery.hpp>
#include <anna/comm/Resource.hpp>
#include <anna/comm/Delivery.hpp>

using namespace std;
using namespace anna;

const Millisecond comm::Delivery::DefaultRecoveryTime(60000);

void comm::Delivery::initialize()
throw(RuntimeException) {
  LOGMETHOD(TraceMethod tm("comm::Delivery", "initialize", ANNA_FILE_LOCATION));
// No es necesario protegerlo porque cuando se ejecuta todava estarmos en la parte
// en la que no se han lanzado los thread's
//   Guard guard (*this, "comm::Delivery::initialize");
  do_initialize();

  if(a_resources.empty() == true) {
    string msg(asString());
    msg += " | No resources assigned";
    Logger::warning(msg, ANNA_FILE_LOCATION);
  }

  a_isAvailable = isAvailable();
  LOGDEBUG(
    string msg("comm::Delivery::initialize | ");
    msg += asString();
    Logger::debug(msg, ANNA_FILE_LOCATION);
  );
}

//---------------------------------------------------------------------------------------------
// Selecciona alguno de los recursos que tiene asociados este servicio de reparto.
//
// (1) Cada cierto tiempo comprueba si hay que habilitar alguno de los recursos que han sido
// desactivados con motivo de errores internos. Los que hayan sido desactivados por motivos
// externos seran recuperados por el comm::ConnectionRecover, que se estara encargando de
// intentar reconectar cada cierto tiempo.
//
// (2) En principio, supone que va a recuperar todos los recursos con problemas.
//     (2.1) Si queda algun recurso con problemas, activa el timeStamp para seguir comprobando
//           los recursos con fallos internos.
//---------------------------------------------------------------------------------------------
comm::Resource* comm::Delivery::apply()
throw(RuntimeException) {
  LOGMETHOD(TraceMethod tm(Logger::Local7, "comm::Delivery", "apply", ANNA_FILE_LOCATION));
  Guard guard(*this, "comm::Delivery::apply");

  if(a_timeStamp != 0) {                                      // (1)
    const Millisecond now = functions::millisecond();

    if((now - a_timeStamp) > (a_recoveryTime >> 1)) {
      comm::Resource* resource;
      a_timeStamp = 0;                                     // (2)

      for(iterator ii = begin(), maxii = end(); ii != maxii; ii ++) {
        resource = Delivery::resource(ii);

        if(internalErrorDetected(resource) == false)
          continue;

        if((now - resource->getTimeStamp()) > a_recoveryTime)
          unsafe_recover(resource);
        else if(a_timeStamp == 0)
          a_timeStamp = now;                           // (2.1)
      }
    }
  }

  comm::Resource* result = do_apply();

  if(result == NULL) {
    string msg(asString());
    msg += " | No resources available";
    throw RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  LOGDEBUG(
    string msg("comm::Delivery::apply | ");
    msg += asString();
    msg += " | ";
    msg += result->asString();
    Logger::debug(msg, ANNA_FILE_LOCATION)
  );
  return result;
}

bool comm::Delivery::fault(const Resource* resource)
throw() {
  LOGMETHOD(TraceMethod tm(Logger::Local7, "comm::Delivery", "fault", ANNA_FILE_LOCATION));

  if(resource == NULL)
    return !a_isAvailable;

  Guard guard(*this, "comm::Delivery::fault");

  if(find(comm::Delivery::begin(), end(), resource) == end())                // (1)
    return !a_isAvailable;

  const bool result = do_fault(resource);                                    // (2)

  if(a_timeStamp == 0 && internalErrorDetected(resource) == true)
    a_timeStamp = functions::millisecond();

  const int aux = a_isAvailable;
  a_isAvailable = !result;
  LOGWARNING(
    string msg("comm::Delivery::fault | ");
    msg += asString();
    msg += " | ";
    msg += resource->asString();
    Logger::warning(msg, ANNA_FILE_LOCATION)
  );

  if(a_isAvailable == false && a_isAvailable != aux) {
    string msg("comm::Delivery::fault | ");
    msg += asString();
    Logger::error(msg, ANNA_FILE_LOCATION);
  }

  return result;
}

bool comm::Delivery::recover(const Resource* resource)
throw() {
  if(resource == NULL)
    return a_isAvailable;

  LOGMETHOD(TraceMethod tm(Logger::Local7, "comm::Delivery", "recover", ANNA_FILE_LOCATION));
  Guard guard(*this, "comm::Delivery::recover");
  return unsafe_recover(resource);
}

bool comm::Delivery::unsafe_recover(const Resource* resource)
throw() {
  LOGMETHOD(TraceMethod tm(Logger::Local7, "comm::Delivery", "unsafe_recover", ANNA_FILE_LOCATION));

  if(find(comm::Delivery::begin(), end(), resource) == end())                // (1)
    return a_isAvailable;

  a_isAvailable = do_recover(resource);
  LOGWARNING(
    string msg("comm::Delivery::recover | ");
    msg += asString();
    msg += " | ";
    msg += resource->asString();
    Logger::warning(msg, ANNA_FILE_LOCATION)
  );
  return a_isAvailable;
}

bool comm::Delivery::contains(const Resource* resource) const
throw() {
  if(resource == NULL)
    return false;

  Guard guard(*this, "comm::Delivery::contains");
  return do_contains(resource);
}

// Devolver true si no tiene disponible ninguno de los recursos asociados.
bool comm::Delivery::do_fault(const Resource* resource)
throw() {
  return (isAvailable() == true) ? false : true;
}

// Devolver true si tiene disponible alguno de los recursos que tiene asociados.
bool comm::Delivery::do_recover(const Resource* resource)
throw() {
  return isAvailable();
}

//-------------------------------------------------------------------------------------------
// Un recurso puede estar desactivado porque el Communicator a detectado una rotura de
// conexion, lo que consideraremos un error externo, o porque se han detectado errores al
// intentar enviar, lo que consideraremos un error interno.
//
//-------------------------------------------------------------------------------------------
bool comm::Delivery::isAvailable() const
throw() {
  bool result(false);

  for(const_iterator ii = begin(), maxii = end(); ii != maxii; ii ++) {
    const comm::Resource* resource = Delivery::resource(ii);

    if(resource->isEnabled() == true && resource->isAvailable() == true) {
      result = true;
      break;
    }
  }

  return result;
}

void comm::Delivery::add(Resource* resource)
throw(RuntimeException) {
  if(resource == NULL) {
    string msg(asString());
    msg += " | Cannot attach a NULL resource";
    throw RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  Guard guard(*this, "comm::Delivery::add");

  if(do_contains(resource) == true)
    return;

  LOGINFORMATION(
    string msg("comm::Delivery::add | ");
    msg += asString();
    msg += " | ";
    msg += resource->asString();
    Logger::information(msg, ANNA_FILE_LOCATION);
  );
  a_resources.push_back(resource);
}

bool comm::Delivery::do_contains(const comm::Resource* resource) const
throw() {
  for(const_iterator ii = begin(), maxii = end(); ii != maxii; ii ++)
    if(*Delivery::resource(ii) == *resource)
      return true;

  return false;
}

string comm::Delivery::asString() const
throw() {
  string result("comm::Delivery { Name: ");
  result += a_name;
  result += functions::asText(" | Available: ", a_isAvailable);
  result += " | N: ";
  result += functions::asString(a_resources.size());

  if(a_timeStamp != 0)
    result += " | InternalError: Detected";

  return result += " }";
}

xml::Node* comm::Delivery::asXML(xml::Node* parent) const
throw() {
  xml::Node* node = parent->createChild("comm.Delivery");
  node->createAttribute("Name", a_name);
  node->createAttribute("Available", functions::asString(a_isAvailable));

  if(a_timeStamp != 0)
    node->createAttribute("InternalError", "Detected");

  xml::Node* resources = node->createChild("comm.Resources");

  for(const_iterator ii = begin(), maxii = end(); ii != maxii; ii ++) {
    const Resource* commResource = resource(ii);
    commResource->asAttribute(resources->createChild("comm.Resource"));
  }

  return node;
}

/*
 * Devuelve true si el recurso recibido como parametro ha sido desactivado como
 * resultado de un error interno, es decir, que se ha intentado enviar un mensaje
 * y se ha obtenido una excepcion (EAGAIN, EPIPE, etc, etc), o false en otro caso.
 *
 * Si el 'isAvailable' fuera 'false' => que la conexion con el servidor remoto
 * esta rota, y por tanto, en caso de que este marcado como de auto-reconexion
 * estara siendo recuperada por el comm::ConnectionRecover.
 */
/*static*/
bool comm::Delivery::internalErrorDetected(const Resource* resource)
throw() {
  return resource->isEnabled() == false && resource->isAvailable() == true;
}

