// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <sys/time.h>
#include <fcntl.h>
#include <strings.h>
#include <sys/types.h>
#include <poll.h>

#include <algorithm>

#include <anna/core/tracing/Logger.hpp>
#include <anna/core/tracing/TraceMethod.hpp>
#include <anna/core/DataBlock.hpp>

#include <anna/xml/Node.hpp>

#include <anna/comm/Socket.hpp>
#include <anna/comm/INetAddress.hpp>
#include <anna/comm/functions.hpp>
#include <anna/comm/Application.hpp>
#include <anna/comm/TransportFactory.hpp>
#include <anna/comm/ReceiverFactory.hpp>

using namespace std;
using namespace anna;
using namespace anna::comm;

Socket::Socket(const Domain::_v domain, const Type::_v type, TransportFactory* transportFactory) :
  a_domain(domain),
  a_type(type),
  a_fd(-1),
  a_isBound(false),
  a_reuseMode(false),
  a_category(0),
  a_receiverFactory(NULL) {
  if((a_transportFactory = transportFactory) == NULL) {
    try {
      comm::Application& app = comm::functions::getApp();
      a_transportFactory = &app.comm::Application::getDefaultTransportFactory();
    } catch(Exception& ex) {
      ex.trace();
    }
  }
}

Socket::Socket(const INetAddress& localAddress, const Type::_v type, TransportFactory* transportFactory) :
  a_domain(Domain::Inet),
  a_type(type),
  a_fd(-1),
  a_localAccessPoint(localAddress),
  a_isBound(false),
  a_reuseMode(false),
  a_category(0),
  a_receiverFactory(NULL) {
  if((a_transportFactory = transportFactory) == NULL) {
    try {
      comm::Application& app = comm::functions::getApp();
      a_transportFactory = &app.comm::Application::getDefaultTransportFactory();
    } catch(Exception& ex) {
      ex.trace();
    }
  }
}

Socket::Socket(const std::string& path, const Type::_v type, TransportFactory* transportFactory) :
  a_transportFactory(NULL),
  a_domain(Domain::Unix),
  a_type(type),
  a_fd(-1),
  a_localAccessPoint(path),
  a_isBound(false),
  a_reuseMode(false),
  a_category(0),
  a_receiverFactory(NULL) {
  if((a_transportFactory = transportFactory) == NULL) {
    try {
      comm::Application& app = comm::functions::getApp();
      a_transportFactory = &app.comm::Application::getDefaultTransportFactory();
    } catch(Exception& ex) {
      ex.trace();
    }
  }
}

Socket::~Socket() {
  try {
    close();
  } catch(RuntimeException& ex) {
    ex.trace();
  }
}

bool Socket::support(const char* transportClassName) const
throw() {
  if(a_transportFactory == NULL)
    return false;

  return anna_strcmp(a_transportFactory->getName().c_str(), transportClassName) == 0;
}

//---------------------------------------------------------------------------
// No hace falta que sea MT-safe porque siempre se invoca desde m�odo que
// ya tienen activa su propia zona de exclusin.
//---------------------------------------------------------------------------
void Socket::open()
throw(RuntimeException) {
  if(a_fd != -1)
    return;

  if(a_transportFactory == NULL) {
    std::string msg(asString());
    msg += " | Transport factory was not especified";
    throw RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  anna_comm_socket_check(a_fd = socket((int) a_domain, (int) a_type, 0), "Cannot create the socket");
  int value;
  socklen_t len = sizeof(int);
  anna_comm_socket_check(
    getsockopt(a_fd, SOL_SOCKET, SO_REUSEADDR, &value, &len),
    "Cannot obtain reuse mode"
  );
  a_reuseMode = (value == 1);

  if(a_type == Type::Datagram) {
    value = 1;
    anna_comm_socket_check(
      setsockopt(a_fd, SOL_SOCKET, SO_BROADCAST, &value, sizeof(int)),
      "Cannot activate broadcast system"
    );
  }

  LOGDEBUG(
    string msg("comm::Socket::open | ");
    msg += asString();
    Logger::debug(msg, ANNA_FILE_LOCATION);
  );
}

void Socket::close()
throw() {
  if(a_fd == -1)
    return;

  Guard guard(*this, "comm::Socket::close");

  if(a_fd != -1) {
    a_isBound = false;
    do_close();
    a_fd = -1;
  }

  LOGDEBUG(
    string msg("comm::Socket::close | ");
    msg += asString();
    Logger::debug(msg, ANNA_FILE_LOCATION);
  );
}

bool Socket::setBlockingMode(const bool blockingMode)
throw(RuntimeException) {
  int opts;
  int result;
  anna_socket_assert(a_fd == -1, "Socket is not opened");
  anna_comm_socket_check(opts = fcntl(a_fd, F_GETFL), "Cannot obtain state bits (F_GETFL)");
  result = !(opts & O_NONBLOCK);

  if(blockingMode == false)
    opts |= O_NONBLOCK;
  else
    opts &= ~O_NONBLOCK;

  anna_comm_socket_check(fcntl(a_fd, F_SETFL, opts), "Cannot set state bits (F_SETFL)");
  LOGDEBUG(
    string msg("comm::Socket::setBlockingMode | Current: ");
    msg += functions::asString(blockingMode);
    msg += functions::asText(" | Previous: ", result != 0);
    Logger::debug(msg, ANNA_FILE_LOCATION);
  )
  return (result != 0);
}

bool Socket::setReuseMode(const bool reuseMode)
throw(RuntimeException) {
  anna_socket_assert(a_fd == -1, "Socket is not opened");

  if(a_reuseMode == reuseMode)
    return reuseMode;

  bool result;
  result = a_reuseMode;
  int value = htons((a_reuseMode = reuseMode) == true);
  anna_comm_socket_check(
    setsockopt(a_fd, SOL_SOCKET, SO_REUSEADDR, &value, sizeof(int)),
    "Cannot set the reuse mode"
  );
  return result;
}

void Socket::bind()
throw(RuntimeException) {
  Guard guard(*this, "comm::Socket::bind");

  if(a_fd == -1)
    open();

  anna_socket_assert(a_isBound == true, "Already attached");
  setReuseMode(true);
  sockaddr* s(NULL);
  int len(0);
  a_localAccessPoint.translate(*this, s, len);

  if(do_bind(s, len) < 0) {
    RuntimeException ex(asString(), errno, ANNA_FILE_LOCATION);
    close();
    throw ex;
  }

  a_isBound = true;
  LOGDEBUG(
    string msg("comm::Socket::bind | ");
    msg += asString();
    Logger::debug(msg, ANNA_FILE_LOCATION);
  );
}

int Socket::do_bind(const struct sockaddr *addr, const int len)
throw(RuntimeException) {
  return ::bind(a_fd, addr, len);
}

std::string Socket::asString() const
throw() {
  std::string msg("comm::Socket { Domain: ");

  if(this == NULL)
    return msg += " <null> }";

  switch(a_domain) {
  case Domain::Unix: msg += "Unix"; break;
  case Domain::Inet: msg += "Inet"; break;
  default: msg += '<'; msg += functions::asString((int) a_domain); msg += '>'; break;
  }

  msg += " | Type: ";
  msg += (a_type == Type::Stream) ? "Stream" : "Datagram";
  msg += " | ";

  if(a_transportFactory != NULL) {
    msg += a_transportFactory->asString();
    msg += " | ";
  }

  msg += "fd: ";
  msg += functions::asString(a_fd);
  msg += " | LocalPoint: ";
  a_localAccessPoint.asString(msg);
  msg += " | Bound: ";
  msg += functions::asString(a_isBound);
  msg += " | Opened: ";
  msg += functions::asString(isOpened());

  if(a_category != 0)
    msg += functions::asText(" | Category: ", a_category);

  if(a_receiverFactory != NULL) {
    msg += " | ";
    msg += a_receiverFactory->asString();
  }

  return msg += " }";
}

xml::Node* Socket::asXML(xml::Node* parent) const
throw(RuntimeException) {
  const char* aux;
  xml::Node* result = parent->createChild("comm.Socket");

  switch(a_domain) {
  case Domain::Unix: aux = "Unix"; break;
  case Domain::Inet: aux = "Inet"; break;
  default: aux = functions::asString((int) a_domain).c_str(); break;
  }

  result->createAttribute("Domain", aux);
  result->createAttribute("Type", (a_type == Type::Stream) ? "Stream" : "Datagram");

  if(a_transportFactory != NULL)
    a_transportFactory->asXML(result);

  result->createAttribute("fd", a_fd);
  result->createAttribute("Bound", functions::asString(a_isBound));
  result->createAttribute("Opened", functions::asString(isOpened()));
  a_localAccessPoint.asXML("comm.LocalPoint", result);

  if(a_fd != -1) {
    int opts;

    if((opts = fcntl(a_fd, F_GETFL)) != -1)
      result->createAttribute("Mode", (opts & O_NONBLOCK) ? "Non-Blocking" : "Blocking");
  }

  if(a_category != 0)
    result->createAttribute("Category", a_category);

  if(a_receiverFactory != NULL)
    a_receiverFactory->asXML(result);

  return result;
}

const char* Socket::asText(const Socket::Notify::_v v)
throw() {
  static const char* text [] = { "None", "ReceiveData", "Close", "Corrupt" };
  return text [v];
}

