// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <anna/diameter.comm/Server.hpp>
#include <anna/diameter.comm/Entity.hpp>
#include <anna/diameter.comm/Engine.hpp>
#include <anna/diameter.comm/OamModule.hpp>
#include <anna/diameter/codec/functions.hpp>
#include <anna/diameter/helpers/base/defines.hpp>
#include <anna/core/functions.hpp>
#include <anna/time/functions.hpp>
#include <anna/diameter/helpers/base/functions.hpp>
#include <anna/diameter/helpers/dcca/functions.hpp>

#include <anna/core/tracing/Logger.hpp>
#include <anna/core/functions.hpp>
#include <anna/core/tracing/TraceMethod.hpp>


using namespace anna;
using namespace anna::diameter;
using namespace anna::diameter::comm;



void Entity::initialize() throw() {
  a_engine = NULL;
  a_servers.clear(); // importante (el recycler creo que no lo tocaba)
  a_available = false;
  a_deprecated = false;
  a_socketListLiteral = "";
  a_primarySocketLiteral = "";
  a_secondarySocketLiteral = "";
  a_description = "";
  a_category = 0;
  a_lastUsedResource = NULL;
  a_balance = false;
  a_sessionBasedModelsType = SessionBasedModelsType::SessionIdLowPart;
}


void Entity::assertReady() throw(anna::RuntimeException) {
  if(a_servers.size() != a_maxServers) {
    std::string msg(asString());
    msg += " | Non-configured entity: you must add the remaining servers before any operation (bind, send, etc.)";
    throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
  }
}


void Entity::addServer(const socket_t & serverId)
throw(anna::RuntimeException) {
  if(a_servers.size() == a_maxServers) {
    LOGDEBUG
    (
      std::string msg = "Maximum number of servers reached for this entity (";
      msg += anna::functions::asString(a_maxServers);
      msg += "). Please use clear()/close() primitives";
      anna::Logger::debug(msg, ANNA_FILE_LOCATION);
    );
    return;
  }

  if(!a_engine)
    throw anna::RuntimeException("Invalid engine reference (NULL)", ANNA_FILE_LOCATION);

  Server *s = a_engine->createServer(this, serverId);
  a_servers.push_back(s);
  a_deliveryIterator = begin();
}


int Entity::readSocketId(const Message* message, int maxClientSessions) const throw() {

  if(a_sessionBasedModelsType == SessionBasedModelsType::RoundRobin) return -1;  // IEC also would return -1

  try {
    // Service-Context-Id:
    anna::diameter::helpers::dcca::ChargingContext::_v chargingContext;
    std::string scid = anna::diameter::helpers::dcca::functions::getServiceContextId(message->getBody(), chargingContext);

    switch(chargingContext) {
    case anna::diameter::helpers::dcca::ChargingContext::Data:
    case anna::diameter::helpers::dcca::ChargingContext::Voice:
    case anna::diameter::helpers::dcca::ChargingContext::Content: {
      // Session-Id: '<DiameterIdentity>;<high 32 bits>;<low 32 bits>[;<optional value>="">]'
      std::string sid = anna::diameter::helpers::base::functions::getSessionId(message->getBody());
      std::string diameterIdentity, optional;
      anna::U32 high, low;
      anna::diameter::helpers::base::functions::decodeSessionId(sid, diameterIdentity, high, low /* context-teid */, optional);

      if(a_sessionBasedModelsType == SessionBasedModelsType::SessionIdLowPart) return (low % maxClientSessions);

      if(a_sessionBasedModelsType == SessionBasedModelsType::SessionIdHighPart) return (high % maxClientSessions);

      if(a_sessionBasedModelsType == SessionBasedModelsType::SessionIdOptionalPart) return (atoi(optional.c_str()) % maxClientSessions);
    }
    case anna::diameter::helpers::dcca::ChargingContext::SMS:
    case anna::diameter::helpers::dcca::ChargingContext::MMS:
    case anna::diameter::helpers::dcca::ChargingContext::Unknown:
    default:
       return -1; // IEC model and Unknown traffic types
    }
  } catch(anna::RuntimeException &ex) {
    LOGDEBUG(
      std::string msg = ex.getText();
      msg += " | Round-robin between sessions will be used to send";
      anna::Logger::debug(msg, ANNA_FILE_LOCATION);
    );
  }

  return -1;
}


bool Entity::send(const Message* message) throw(anna::RuntimeException) {
  LOGMETHOD(anna::TraceMethod tttm("diameter::comm::Entity", "send", ANNA_FILE_LOCATION));
  assertReady();
  // Carried socket id (forwarding/proxy features):
  std::string carriedSocketId = message->getRequestClientSessionKey();

  if(carriedSocketId != "") {
    // Send (it was a request, we forward the answer):
    try {
      ClientSession * fixedClientSession = a_engine->findClientSession(carriedSocketId); // exception if not found
      fixedClientSession->send(message);
      return true;
    } catch(anna::RuntimeException &ex) {
      std::string msg = "Cannot deliver answer through a fixed client session (";
      msg += carriedSocketId;
      msg += "). Perhaps it existed but not now. Ignore";
      anna::Logger::error(msg, ANNA_FILE_LOCATION);
      ex.trace();
    }
  }

  ////////////////////////////////////////////////////////////////////////////////////////
  // BALANCE vs STANDARD
  // Balance algorythm remember last delivery resource used, balancing from start to end,
  // understanding start as next resource to last used, and end as last used. Standard
  // algorithm always starts at primary (first defined) server.
  ////////////////////////////////////////////////////////////////////////////////////////

  // Balance
  if(a_balance) {
    for(int k = 0; k < getMaxServers(); k++) {   // try round-robin only over one cycle,
      // no matter where you are: don't repeat same server
      if(a_deliveryIterator == end()) a_deliveryIterator = begin();

      a_lastUsedResource = (*a_deliveryIterator);
      a_deliveryIterator++;
      // At 'readSocketId' documentation:
      //      If server is configured as single session (max client sessions equal to 1), entity will ignore
      //      this method because it won't affect the session selection.
      int serverSessions = a_lastUsedResource->getMaxClientSessions();
      int socketId = (serverSessions > 1) ? readSocketId(message, serverSessions) : -1; // optimization

      if(a_lastUsedResource->send(message, socketId))  // exception only possible at findClientSession within server send procedure
        return true;
    }
  } else {
    // Standard (no balance) //   start at begining, try secondary, and so on until end.
    std::vector<Server*>::iterator it = begin();

    while(it != end()) {
      a_lastUsedResource = (*it);
      it++;
      // At 'readSocketId' documentation:
      //      If server is configured as single session (max client sessions equal to 1), entity will ignore
      //      this method because it won't affect the session selection.
      int serverSessions = a_lastUsedResource->getMaxClientSessions();
      int socketId = (serverSessions > 1) ? readSocketId(message, serverSessions) : -1; // optimization

      if(a_lastUsedResource->send(message, socketId))  // exception only possible at findClientSession within server send procedure
        return true;
    }
  }

  // END BALANCE AND TRY ALGORITHM or STANDARD ///////////////////////////////////////////
  ////////////////////////////////////////////////////////////////////////////////////////
  // Here, sent has failed:
  // OAM
  OamModule &oamModule = OamModule::instantiate();

  if(a_maxServers != 2) {
    oamModule.activateAlarm(OamModule::Alarm::UnableToDeliverDiameterMessageToEntityDefinedAs__s__, a_socketListLiteral.c_str());
    oamModule.count(OamModule::Counter::UnableToDeliverOverEntity);
  } else {
    OamModule &oamModule = OamModule::instantiate();
    oamModule.activateAlarm(OamModule::Alarm::UnableToDeliverDiameterMessageToEntityDefinedAsPrimary__s__AndSecondary__s__,
                            a_primarySocketLiteral.c_str(),
                            a_secondarySocketLiteral.c_str());
    oamModule.count(OamModule::Counter::UnableToDeliverOverEntity);
  }

  return false;
}

bool Entity::broadcast(const Message* message) throw(anna::RuntimeException) {
  LOGMETHOD(anna::TraceMethod tttm("diameter::comm::Entity", "broadcast", ANNA_FILE_LOCATION));
  assertReady();
  bool allok = true;
  bool ok;

  for(std::vector<Server*>::iterator it = begin(); it != end(); it++) {
    try {
      ok = (*it)->broadcast(message);

      if(!ok) allok = false;
    } catch(anna::RuntimeException &ex) {
      ex.trace();
      allok = false;
    }
  }

  return allok;
}

bool Entity::bind() throw(anna::RuntimeException) {
  LOGMETHOD(anna::TraceMethod tttm("diameter::comm::Entity", "bind", ANNA_FILE_LOCATION));
  assertReady();
  bool result = true; // all OK return

  for(std::vector<Server*>::iterator it = begin(); it != end(); it++) {
    try {
      (*it)->bind();
    } catch(anna::RuntimeException &ex) {
      ex.trace();
      result = false;
    }
  }

  return result;
}

void Entity::raiseAutoRecovery(bool autoRecovery) throw(anna::RuntimeException) {
  LOGMETHOD(anna::TraceMethod tttm("diameter::comm::Entity", "raiseAutoRecovery", ANNA_FILE_LOCATION));
  assertReady();

  for(std::vector<Server*>::iterator it = begin(); it != end(); it++)
    (*it)->raiseAutoRecovery(autoRecovery);
}

void Entity::setClassCodeTimeout(const ClassCode::_v v, const anna::Millisecond & millisecond) throw() {
  LOGMETHOD(anna::TraceMethod tttm("diameter::comm::Entity", "setClassCodeTimeout", ANNA_FILE_LOCATION));
  assertReady();

  for(std::vector<Server*>::iterator it = begin(); it != end(); it++) {
    try {
      (*it)->setClassCodeTimeout(v, millisecond);
    } catch(anna::RuntimeException &ex) {
      ex.trace();
    }
  }
}


// Private close/destroy method
void Entity::close(bool destroy) throw(anna::RuntimeException) {
  LOGMETHOD(anna::TraceMethod tttm("diameter::comm::Entity", "close", ANNA_FILE_LOCATION));

  if(!a_engine)
    throw anna::RuntimeException("Invalid engine reference (NULL)", ANNA_FILE_LOCATION);

  assertReady();

  for(std::vector<Server*>::iterator it = begin(); it != end(); it++)
    a_engine->closeServer(*it, destroy);
}

const char* Entity::asText(const SessionBasedModelsType::_v sbmt)
throw() {
  static const char* text [] = { "RoundRobin", "SessionIdOptionalPart", "SessionIdHighPart", "SessionIdLowPart" };
  return text [sbmt];
}

socket_v Entity::getAddressPortList() const throw() {
  socket_v result;

  for(std::vector<Server*>::const_iterator it = begin(); it != end(); it++) {
    socket_t address((*it)->getAddress(), (*it)->getPort());
    result.push_back(address);
  }

  return result;
}

int Entity::getOTARequests() const throw() {
  int result = 0;

  for(std::vector<Server*>::const_iterator it = begin(); it != end(); it++)
    result += (*it)->getOTARequests();

  return result;
}

void Entity::childIdle() const throw() {
  // Check father engine idleness:
  if(idle()) a_engine->eraseDeprecatedIdleEntities();
}


void Entity::hide() throw() {
  for(std::vector<Server*>::iterator it = begin(); it != end(); it++)
    (*it)->hide();
}

void Entity::show() throw() {
  for(std::vector<Server*>::iterator it = begin(); it != end(); it++)
    (*it)->show();
}

bool Entity::hidden() const throw() {
  for(std::vector<Server*>::const_iterator it = begin(); it != end(); it++)
    if((*it)->shown()) return false;

  return true;
}
bool Entity::shown() const throw() {
  for(std::vector<Server*>::const_iterator it = begin(); it != end(); it++)
    if((*it)->hidden()) return false;

  return true;
}

void Entity::eventPeerShutdown(const ClientSession* clientSession) throw() {
  LOGWARNING(
    std::string msg(clientSession->asString());
    msg += " | eventPeerShutdown";
    anna::Logger::warning(msg, ANNA_FILE_LOCATION);
  );
}

void Entity::eventRequestRetransmission(const ClientSession* clientSession, Message *request) throw() {
  LOGWARNING(
    std::string msg(clientSession->asString());

    HopByHop hopByHop = codec::functions::getHopByHop(request->getBody()); // context identification
    int retries = request->getRetries();

    msg += anna::functions::asString(" | eventRequestRetransmission: request with application HopByHop: %u; remaining %d retries", hopByHop, retries);
    anna::Logger::warning(msg, ANNA_FILE_LOCATION);
  );
}

std::string Entity::asString() const throw() {
  std::string result("diameter::comm::Entity { ");
  std::string originRealm = a_engine->getOriginRealmName();
  std::string originHost = a_engine->getOriginHostName();

  result += "Parent Engine Origin-Realm: ";
  result += (originRealm != "") ? originRealm:"[not configured]";
  result += " | Parent Engine Origin-Host: ";
  result += (originHost != "") ? originHost:"[not configured]";

  result += " | Category: ";
  result += anna::functions::asString(a_category);

  if(a_description != "") {
    result += " | Description: '";
    result += a_description;
  }

  result += " | Available: ";
  result += a_available ? "yes" : "no";
  result += " | Deprecated: ";
  result += a_deprecated ? "yes" : "no";
  result += " | Max servers supported: ";
  result += anna::functions::asString(a_maxServers);
  result += " | Currently configured servers: ";
  result += anna::functions::asString(a_servers.size());
  result += anna::functions::asString(" | OTA requests: %d%s", getOTARequests(), idle() ? " (idle)" : "");
  result += " | Last Incoming Activity Time: ";
  result += a_lastIncomingActivityTime.asString();
  result += " | Last Outgoing Activity Time: ";
  result += a_lastOutgoingActivityTime.asString();
  result += " | Hidden: ";
  result += (hidden() ? "yes" : "no");
  result += " | SessionBasedModelsType: ";
  result += asText(a_sessionBasedModelsType);
  result += "\n";

  for(std::vector<Server*>::const_iterator it = begin(); it != end(); it++) {
    result += "\n";
    result += (*it)->asString();
  }

  return result;
}

anna::xml::Node* Entity::asXML(anna::xml::Node* parent) const throw() {
  anna::xml::Node* result = parent->createChild("diameter.Entity");
  std::string originRealm = a_engine->getOriginRealmName();
  std::string originHost = a_engine->getOriginHostName();

  if(originRealm != "") result->createAttribute("ParentEngineOriginRealm", originRealm);
  if(originHost != "") result->createAttribute("ParentEngineOriginHost", originHost);

  result->createAttribute("Category", anna::functions::asString(a_category));

  if(a_description != "") result->createAttribute("Description", a_description);

  result->createAttribute("Available", a_available ? "yes" : "no");
  result->createAttribute("Deprecated", a_deprecated ? "yes" : "no");
  result->createAttribute("MaxServersSupported", anna::functions::asString(a_maxServers));
  result->createAttribute("CurrentlyConfiguredServers", anna::functions::asString(a_servers.size()));
  result->createAttribute("OTArequests", anna::functions::asString("%d%s", getOTARequests(), idle() ? " (idle)" : ""));
  result->createAttribute("LastIncomingActivityTime", a_lastIncomingActivityTime.asString());
  result->createAttribute("LastOutgoingActivityTime", a_lastOutgoingActivityTime.asString());
  result->createAttribute("Hidden", hidden() ? "yes" : "no");
  result->createAttribute("SessionBasedModelsType:", asText(a_sessionBasedModelsType));
  anna::xml::Node* servers = result->createChild("Entity.Servers");

  for(std::vector<Server*>::const_iterator it = begin(); it != end(); it++)
    (*it)->asXML(servers);

  return result;
}

void Entity::availabilityLost() throw() {
  a_available = false;
  LOGDEBUG(
    std::string msg = "diameter::comm::Entity { Description: ";
    msg += getDescription();
    msg += " } has lost its availability";
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );
  OamModule &oamModule = OamModule::instantiate();

  if(a_maxServers != 2) {
    oamModule.activateAlarm(OamModule::Alarm::c_LostAvailabilityOverEntityDefinedAs__s__, a_socketListLiteral.c_str());
    oamModule.count(OamModule::Counter::LostAvailabilityOverEntity);
  } else {
    oamModule.activateAlarm(OamModule::Alarm::c_LostAvailabilityOverEntityDefinedAsPrimary__s__AndSecondary__s__,
                            a_primarySocketLiteral.c_str(),
                            a_secondarySocketLiteral.c_str());
    oamModule.count(OamModule::Counter::LostAvailabilityOverEntity);
  }

  a_engine->availabilityLost(this);
  a_engine->refreshAvailabilityForEntities();
}


void Entity::availabilityRecovered() throw() {
  a_available = true;
  LOGDEBUG(
    std::string msg = "diameter::comm::Entity { Description: ";
    msg += getDescription();
    msg += " } has recovered its availability";
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );
  OamModule &oamModule = OamModule::instantiate();

  if(a_maxServers != 2) {
    oamModule.cancelAlarm(OamModule::Alarm::c_LostAvailabilityOverEntityDefinedAs__s__, a_socketListLiteral.c_str());
    oamModule.count(OamModule::Counter::RecoveredAvailabilityOverEntity);
  } else {
    oamModule.cancelAlarm(OamModule::Alarm::c_LostAvailabilityOverEntityDefinedAsPrimary__s__AndSecondary__s__,
                          a_primarySocketLiteral.c_str(),
                          a_secondarySocketLiteral.c_str());
    oamModule.count(OamModule::Counter::RecoveredAvailabilityOverEntity);
  }

  a_engine->availabilityRecovered(this);
  a_engine->refreshAvailabilityForEntities();
}


bool Entity::refreshAvailability() throw() {
  // Here available
  if(a_available) {  // check not-bound state for all servers:
    bool isolate = true;

    for(std::vector<Server*>::const_iterator it = begin(); it != end(); it++)
      if((*it)->isAvailable()) { isolate = false; break; }

    if(isolate) {
      availabilityLost();
      return true;
    }

    return false;
  }

  // Here not available
  for(std::vector<Server*>::const_iterator it = begin(); it != end(); it++)
    if((*it)->isAvailable()) {
      availabilityRecovered();
      return true;
    }

  return false;
}

//------------------------------------------------------------------------------
//----------------------------------------- Entity::updateIncomingActivityTime()
//------------------------------------------------------------------------------
void Entity::updateIncomingActivityTime() throw() {
  a_lastIncomingActivityTime = anna::functions::millisecond();
  LOGDEBUG
  (
    std::string msg = "Updated INCOMING activity on entity (milliseconds unix): ";
    msg += anna::functions::asString(a_lastIncomingActivityTime.getValue());
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );
}


//------------------------------------------------------------------------------
//----------------------------------------- Entity::updateOutgoingActivityTime()
//------------------------------------------------------------------------------
void Entity::updateOutgoingActivityTime(void) throw() {
  a_lastOutgoingActivityTime = anna::functions::millisecond();
  LOGDEBUG
  (
    std::string msg = "Updated OUTGOING activity on entity (milliseconds unix): ";
    msg += anna::functions::asString(a_lastOutgoingActivityTime.getValue());
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );
}

